from yacht.anchor import *

anchors = {}
_COUNTER = 0

@anchor
def get(req):
    testCookie = getCookie(req)
    print("test cookie")
    
    if "test2" in testCookie:
        print(testCookie["test2"].value)

    return str(_COUNTER)

@anchor
def increment(req):
    global _COUNTER
    _COUNTER += 1
    return str(_COUNTER)

@anchor
def decrement(req):
    global _COUNTER
    _COUNTER -= 1
    cookie = SimpleCookie()
    cookie["test"] = "increment"
    cookie["test2"] = "decrement"
    setCookie(req, cookie)
    return str(_COUNTER)


import functools, json, sys
from http.cookies import SimpleCookie

_anchorModules = {}

def getRequestBody(request):
    reqLength = int(request.headers.get('content-length', 0))
    reqBody = request.rfile.read(reqLength).decode('UTF-8')
    obj = json.loads(reqBody)
    return obj if obj else {}
    
def getCookie(request):
    return SimpleCookie(request.headers.get('Cookie'))

def setCookie(request, cookie):
    for morsel in cookie.values():
        request.send_header("Set-Cookie", morsel.OutputString())

def anchor(func):
    @functools.wraps(func)
    def auto_func(req):
        return func(req)
    global _anchorModules
    moduleName = func.__globals__['__file__'].split('/')[-1][:-3]
    
    if not moduleName in _anchorModules.keys():
        _anchorModules[moduleName] = {}
    _anchorModules[moduleName][func.__name__] = auto_func
    return auto_func

if __name__ == "__main__":
    helpMessage = """
        This module shouldn't be run

        import it into your module like:

        from yacht.anchor import *

        then use @anchor before any function to make it accessible from yacht generated pages
        place your module .py scirpt in the root directory of your page and import it when launching yacht sail 
        by passing it's name in the list after -m 
    """
    print(helpMessage)
    


import sys, re, uuid
try:
    from yaml import load
except ImportError:
    raise RuntimeError("Looks like you haven't installed PyYAML yet")
   
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

indentSpaces = 4

eval_enabled = False

emptyElementTags = [
    "area",
    "base",
    "br",
    "col",
    "embed",
    "hr",
    "img",
    "input",
    "keygen",
    "link",
    "meta",
    "param",
    "source",
    "track",
    "wbr"
]

def uniqueID(prefix):
    return prefix+str(uuid.uuid4())[0:7]
    
def craftHTML(obj, indent):

    repeat = obj.get("repeat", 1)
    
    index = obj.get("index", "$index")
   
    output = ""
    for i in range(repeat):
        for child in obj["craft"]:
            output += f"{' '*indent}{parseHTMLObj(child, indent+indentSpaces)}".replace(index, str(i))
    name = obj.get("name")
    if name is not None:
        output = f"\n{' '*indent}<-- { name or 'crafted segment' } -->${output}\n{' '*indent}<-- end of {name or 'crafted segment' } -->"
    return output

def tugFile(obj, indent, tugStack):
    
    tugID = uniqueID("tug")
    objIterator = iter(obj)
    next(objIterator)
    try:
        output = convertFile(obj["tug"], indent+indentSpaces, tugStack)
    except FileNotFoundError:
        return errorElement(f"file {obj[tag]} was not found")
    token = next(objIterator, None)
    while token is not None:
        output = output.replace(f"${token}", obj[token])
        token = next(objIterator, None)
    output = output.replace(f" id=\"", f" id=\"{tugID}_")
    output = output.replace(f"bindAnchor(\"", f"bindAnchor(\"{tugID}_")
    output = re.sub(r"#(.+) {", fr"#{tugID}_\1 {{", output)
    
    return output
def parseAnchor(obj, indent, id):
    output = ""
    if obj is not None:
        print(obj)
        for anchor in obj:
    # bindAnchor(id, name, args, target, isPolling)
            arguments = str(anchor.get('arguments', "null")).replace("'", '"')
            target = anchor.get('target', "")
            poll = 'true' if anchor.get('poll', False) else 'false'
            trigger = anchor.get('trigger', "")
            redirect = anchor.get('redirect', "")
            output += f"\n{' '*indent}<script>harbour.bindAnchor(\"{id}\", \"{anchor['bind']}\", {arguments}, \"{target}\", {poll}, \"{trigger}\", \"{redirect}\" );</script>"
    return output
def parseHTMLObj(obj, indent, tugStack=[]):
    output = ""
    objIterator = iter(obj)
    tag = next(objIterator) # get first key as tag name
    content = obj[tag]
    if(tag == "craft"):
        return craftHTML(obj, indent)
    elif(tag == "tug"):
        return tugFile(obj, indent, tugStack)
    elif(tag=="script"):
        if(type(content) == str and not '\n' in content):
            return f"\n{' '*indent}<script src=\"{content}\"></script>"
    elif(tag=="style"):
        if(type(content) == str and not '\n' in content):
            return f"\n{' '*indent}<link rel=\"stylesheet\" href=\"{content}\" />"
    output += f"\n{' '*indent}<{tag}"
    # following keys as attributes
    attributes = []
    anchors = None
    anchorsID = ""
    if "anchors" in obj.keys():
            anchors = obj['anchors']
            anchorsID = obj.get("id", uniqueID("anchor"))
            obj["id"] = anchorsID
            objIterator = iter(obj) # recreate the iterator after modifying obj
            next(objIterator) # skip first key
            

    attrTemp = next(objIterator, None)

    while attrTemp is not None:
        output += f" {attrTemp}=\"{obj[attrTemp]}\""
        attrTemp = next(objIterator, None)
    print(anchors)
    if tag in emptyElementTags:
        output += " />"
        output += parseAnchor(anchors, indent, anchorsID)
        return output;
    
    output += ">"
    
    if content is None:
        output += f"</{tag}>"
        output += parseAnchor(anchors, indent, anchorsID)
        return output
    else:
        if(type(content)==str):  
            if '\n' in content:
                if (tag != "script"):
                    content = content.replace('\n', '<br />\n')
                output += f"\n{content}\n{' '*indent}</{tag}>"
            else:
                output += f"{content}</{tag}>"
            output += parseAnchor(anchors, indent, anchorsID)
            return output
        elif(tag == "style"):
            for child in content:
                output += f"{parseCSSObj(child, indent+indentSpaces)}"
            
        else: 
            for child in content:
                output += f"{' '*indent}{parseHTMLObj(child, indent+indentSpaces, tugStack)}"
    output += f"\n{' '*indent}</{tag}>"
    output += parseAnchor(anchors, indent, anchorsID)
    return output

def parseCSSRule(obj, indent):
    output = ""
    objIterator = iter(obj)
    rule = next(objIterator) # get first key as tag name
    output += f"\n{' '*indent}{rule} {'{'}"
    selectors = obj[rule]
    if selectors is not None:
        for selector in selectors:
            output += f"{parseCSSObj(selector, indent+indentSpaces)}"
    output += f"\n{' '*indent}{'}'}"
    return output

def parseCSSObj(obj, indent):
    output = ""
    objIterator = iter(obj)
    selector = next(objIterator) # get first key as tag name
    if selector[0]=="@":
        return parseCSSRule(obj, indent)
    output += f"\n{' '*indent}{selector} {'{'}"
    content = obj[selector]
    if content is None:
        output += "}"
        return output
    else:
        if(type(content)==str):  
            if '\n' in content:
                output += f"\n{content}\n{' '*indent}{'}'}"
            else:
                output += f" {content} {'}'}"
            return output
        else: 
            output += "\n"
            for child in content:
                key = next(iter(child))
                output += f"{' '*(indent+indentSpaces)}{key}: {child[key]};\n"
    output += f"{' '*indent}{'}'}"
    return output

def errorElement(message):
    return f"<b style=\"color:red\">Error:{message}</b>"

def convertFile(filename, indent=0, tugStack=[]):
    output = ""
    print(f"converting file {filename}")
    with open(filename, 'rt', encoding='utf8') as file:
        data = load(file, Loader=Loader)
        root = next(iter(data))
        
        if filename in tugStack:
            return errorElement(f"tugging {filename} from {tugStack[-1]} would cause recursion")
        
        if not tugStack:
            
            if(root == "html"):
                output += "<!DOCTYPE html>\n"
                output += "<!--  Created using YACHT -->\n"
                output += "<!-- Have a very nice day! -->\n"
                tugStack.append(filename)
                output += parseHTMLObj(data, indent, tugStack)
            elif(root == "style" and not len(tugStack)):
                output += "/*  Created using YACHT */\n"
                output += "/* Have a very nice day! */\n"
                tugStack.append(filename)
                for child in data["style"]:
                    output += f"{parseCSSObj(child, indent)}"
            
        elif(root == "html"):
            return errorElement("tugged pages cannot start with html")
        else:
            tugStack.append(filename)
            output += parseHTMLObj(data, indent, tugStack)
        tugStack.pop()    
    return output;


if __name__ == "__main__":
    filename = sys.argv[1] if len(sys.argv)>1 else "input.yaml" 
    print(convertFile(filename))
    


import sys
from yacht.help import describe
from yacht.sail import startServer
from yacht.convert import convertFile
welcomeMessage = """
=== Welcome aboard the yacht! ⛵ ===

"""

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(welcomeMessage)
        print("If you don't know how to get started run again with: yacht.help")
        print("or check out the documentation at https://github.com/Pobulus/yacht")
        print("")
        quit()
    command = sys.argv[1]
    match command:
        case "help":
            print(welcomeMessage)
            if len(sys.argv) == 2:
                print("The following modules can be accessed by running yacht <name> or directly as modules with yacht.<name>")
                print("☸ help - display this help message")
                print("☸ sail - setup a simple http server")
                print("☸ convert - generate html from a yaml file")
                print("")
                print("Type yacht help <name> to learn more about each topic")
                print("")
            else:
                describe(sys.argv[2])
                print("")
        case "sail":
            startServer(sys.argv[2:])
        case "convert":
            print(convertFile(sys.argv[2] if len(sys.argv) > 2 else "input.yaml"))
        case _:
            print(f"Unknown command {command}")
            print(f"Try running with help")
            print("")


from http.server import BaseHTTPRequestHandler, HTTPServer


import time, sys, re, json, os
import importlib, importlib.util
from yacht.convert import convertFile
import yacht.anchor
def createYachtServer(anchorModules):
    class YachtServer(BaseHTTPRequestHandler):
        def detectContentType(self, path):
            ext = re.search(f"\.([^.]+)$", path).group(1)
            print(ext)
            match ext:
                case "yaml" | "yml" | "html" | "htm":
                    return "text/html"
                case "css"|"yass":
                    return "text/css"
                case "js":
                    return "text/javascript"
                case _: 
                    return "text/plain"
        def do_POST(self):
            items = self.path.split('/')
            
            if(len(items) == 4 and items[1] == 'anchor'):
                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                
                
                response = ""
                moduleName = items[2]
                anchorName = items[3]
                try:   
                    response = anchorModules[moduleName].get(anchorName, lambda n: f"{moduleName}:{anchorName} not found")(self)
                    # the anchor can modify headers to alter cookies
                except KeyError:
                    response = f"no such module {moduleName}"
                self.end_headers() 
                self.wfile.write(bytes(response, "utf-8"))
                return YachtServer
            
        def do_GET(self):
            path = "."+self.path;
            if path[-1] == '/':
                path += 'index.yaml'
            print(f"serving file {path}")
            if re.search(r"\.py$", path):
                self.send_response(403)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(bytes(f"<html><body><h1>403</h1><hr/><h3>You don't have permission to view {path}</h3></body></html>", "utf-8"))
                return YachtServer  
            try:
                document = convertFile(path,0,[]) if re.search(r"\.ya..?$", path) else open(path).read()
                self.send_response(200)
                self.send_header("Content-type", self.detectContentType(path))
                self.end_headers()
                self.wfile.write(bytes(document, "utf-8"))

                

            except FileNotFoundError:
                print('File does not exist')
                self.send_response(404)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(bytes(f"<html><body><h1>404</h1><hr/><h3>file {path} was not found</h3></body></html>", "utf-8"))
    return YachtServer        
def startServer(args):
    
    serverAddress = "localhost"
    serverPort = 55555

    
    if len(args):
        print(args)
        for index, arg in enumerate(args):
            if arg[0] == '-':
                try:
                    match arg:
                        case "-p":
                            try:
                                serverPort = int(args[index+1])
                            except ValueError:
                                print(f"{args[index+1]} is not a valid port number")
                        case "-r":
                            rootPath = args[index+1]
                            os.chdir(rootPath)
                            sys.path.append(rootPath)
                            print(sys.path)
                        case "-a":
                            serverAddress = args[index+1]
                        case "-m":
                            moduleNames = args[index+1].split(',')
                            for name in moduleNames:
                                try:
                                    spec = importlib.util.spec_from_file_location(name, f"{os.getcwd()}/{name}.py")
                                    mod = importlib.util.module_from_spec(spec)
                                    spec.loader.exec_module(mod)
                                    print(f" ✅ Successfully loaded [{name}]:")
                                    for anch in yacht.anchor._anchorModules[name].keys():
                                        print(f" ⚓  {name}/{anch}")

                                except ModuleNotFoundError:
                                    print(f" ❌ error loading [{name}]")
                                    print(f"No such module {name}")
                                except FileNotFoundError:
                                    print(f" ❌ error loading [{name}]")
                                    print(f"    No such file: {os.getcwd()}/{name}.py")
                        case _:
                            print(f"Unknown parameter {arg}")
                except IndexError:
                    print(f"missing value for parameter {arg}")
    webServer = HTTPServer((serverAddress, serverPort), createYachtServer(yacht.anchor._anchorModules))
    print("Server started http://%s:%s" % (serverAddress, serverPort))

    try:
        webServer.serve_forever()
    except KeyboardInterrupt:
        pass

    webServer.server_close()
    print("Server stopped.")

if __name__ == "__main__":
    startServer(sys.argv[1:] if len(sys.argv)>1 else []) 



def describe(topic):
    match topic:
        case "sail":
            print("sail lets you create a simple http server to test you yacht page")
            print("available parameters are:")
            print(" ☸  -a ADDR     address for the server")
            print(" ☸  -p PORT     port which the server should be opened")
            print(" ☸  -r PATH     path to your server's root directory")
            print(" ☸  -m MOD1[,MOD2[...]]")
            print("         list of modules with anchors to load")
            print("         they should be .py files in root directory of your website")
        case "convert":
            print("convert a given yaml into html and print the results to stdout")

        case _:
            print(f"No help available for {topic}")

if __name__ == "__main__":
    import sys
    try:
        describe(sys.argv[1])
    except IndexError:
        print("This is a help topic for yacht")
        print("available topics are:")
        print(" ☸  sail")
        print(" ☸  convert")
        print(" ☸  anchor")
        
        



from datetime import datetime
from yacht.anchor import *
import random

@anchor
def time(request):
    return datetime.now().strftime(anchor.getRequestBody(request).get('format', "%H:%M"))

@anchor
def printer(request): 
    print(getRequestBody(request).get("attributes", {}))
    return str("clicked!")

splashes = [
    "Land ahoy!",
    "Drop the anchor!",
    "Set sail!",
    "All aboard!"
]

@anchor
def splash(req):
    return(random.sample(splashes, 1)[0])

